#include <SoftwareSerial.h>
#include "HapticoreLibrary.h"

#define DEBUG
#define UART_DEBUG_BAUDRATE			115200
#define UART_HAPTICORE_BAUDRATE		115200

using namespace HapticoreLibrary;

HapticoreDevice Hapticore;

// Optional: Used to write output on another hardware serial.
static void PrintMessage(String message, bool linebreak);

// Optional: Used to compose and use the operation time of the used controller.
static byte operatingTime[4];
static bool tickEnabled = true;


void setup()
{
#ifdef DEBUG
	Serial3.begin(UART_DEBUG_BAUDRATE);
#endif

	PrintMessage("--- HAPTICORE Arduino Example Project v3.9.0 ---", true);
	Hapticore.AttachDataReceivedHandler(HandleDataReceived);

#ifdef DEBUG
	Hapticore.AttachDebugLogHandler(PrintMessage);
#endif

	// Connect to Hapticore device. Choose hardware serial port &Serial, &Serial1, &Serial2, ...
#ifdef AVR_UNO
	Hapticore.Connect(&Serial, UART_HAPTICORE_BAUDRATE);
#else
	Hapticore.Connect(&Serial1, UART_HAPTICORE_BAUDRATE);
#endif
}

void loop()
{
#ifdef DEBUG
	// Optional: Can be used during development to communicate with the Arduino example project
	ParseInput();
#endif

	// Poll HAPTICORE data
	Hapticore.PollData();
}

// Optional: Output on hardware serial utility function
void PrintMessage(const String message, bool linebreak = true)
{
#ifdef DEBUG
	if (linebreak)
	{
		Serial3.println(message);
	}
	else
	{
		Serial3.print(message);
	}
#endif
}

// Optional: Input to trigger HAPTICORE functionality / read HAPTICORE information
void ParseInput()
{
#ifdef DEBUG
	if (Serial3.available() > 0)
	{
		byte code = Serial3.read();
		
		// Request library version
		if (code == 0x00)
		{
			Hapticore.Request(Register::REG_HAPTICORE_LIBRARY_VERSION);
		}

		// Request the serial number
		else if (code == 0x01)
		{
			Hapticore.Request(Register::REG_HAPTICORE_MANUFACTURER_SERIAL_NUMBER);
			Hapticore.Request(Register::REG_HAPTICORE_ITEM_NUMBER);
		}

		// Request the operating time
		else if (code == 0x02)
		{
			Hapticore.Request(Register::REG_OPERATINGTIME1);
			Hapticore.Request(Register::REG_OPERATINGTIME2);
		}

		// Send a new value for the tick angle
		else if (code == 0xA0)
		{
			RegisterData data;
			data.FloatType = 3.0f;
			Hapticore.Send(Register::REG_TICK_ANGLE_CW, data);
		}
		
		// Send a different value for the tick angle
		else if (code == 0xA1)
		{
			RegisterData data;
			data.FloatType = 15.0f;
			Hapticore.Send(Register::REG_TICK_ANGLE_CW, data);
		}

		// Toogle tick enabled and disabled
		else if (code == 0xA2)
		{
			RegisterData data;

			if (tickEnabled)
			{
				// Disable
				data.UnsignedType = 0x0;
			}
			else
			{
				// Enable
				data.UnsignedType = 0x1;
			}
			
			Hapticore.Send(Register::REG_TICK_ENABLE, data);
			tickEnabled = !tickEnabled;
		}

		// Enable specific reports
		else if (code == 0xB0)
		{
			RegisterData data;
			data.UnsignedType = static_cast<unsigned short int>((ReportFlags::ENCODER_ANGLE | ReportFlags::ENCODER_MULTI_TURN_COUNT | ReportFlags::TICK_INDEX));
			Hapticore.Send(Register::REG_REPORT_FLAGS, data);
		}

		// Disable "all" reports
		else if (code == 0xB1)
		{
			RegisterData data;

			// Note: It's advisable to keep certain reports always enabled.
			// For example, if the "Device Connection State" report is disabled the library can't detect when the HAPTICORE is dis/re-connected to the control unit.
			data.UnsignedType = static_cast<unsigned short int>((ReportFlags::NOTHING | ReportFlags::DEVICE_CONNECTION_STATE | ReportFlags::DEVICE_ERROR_STATUS));
			Hapticore.Send(Register::REG_REPORT_FLAGS, data);
		}

		else if (code == 0xC0)
		{
			RegisterData data;
			
			data.FloatType = -0.06f;
			Hapticore.Send(Register::REG_CURRENT_CONTROLLER_IDLE_CURRENT, data);

			data.FloatType = 7.0f;
			Hapticore.Send(Register::REG_TICK_ANGLE_CW, data);

			data.FloatType = 0.3f;
			Hapticore.Send(Register::REG_TICK_CURRENT, data);

			data.FloatType = -0.1f;
			Hapticore.Send(Register::REG_TICK_VELOCITY_FACTOR, data);

			data.UnsignedType = static_cast<unsigned short int>(TickMode::TICK_MODE_1);
			Hapticore.Send(Register::REG_TICK_MODE, data);

			data.FloatType = 0.003f;
			Hapticore.Send(Register::REG_TICK_DURATION_MIN, data);

			data.FloatType = 0.007f;
			Hapticore.Send(Register::REG_TICK_DURATION_MAX, data);
		}
	}
#endif
}

// This function is invoked when data is received from the HAPTICORE device.
/* NOTE:
	Values with conversion factors are always stored in 'data.FloatType' (this is the majority of received values).
	There are exceptions though (enabled / disabled values, flags and some other exceptions).
	
	The easiest way to find out how to interpret the 'RegisterData' packet is check the source (HapticoreDevice::HandlePacket(const byte* const packet)) 
	or the documentation 'HAPTICORE Serial Communication Protocol'.

	Here are some examples listed:
*/
void HandleDataReceived(const Register reg, const RegisterData data)
{
	switch (reg)
	{
		case Register::REG_HAPTICORE_MANUFACTURER_SERIAL_NUMBER:
		{
			PrintMessage(String(data.StringType));
			break;
		}

		case Register::REG_HAPTICORE_ITEM_NUMBER:
		{
			PrintMessage(String(data.StringType));
			break;
		}

		case Register::REG_HAPTICORE_LIBRARY_VERSION:
		{
			PrintMessage(String(data.BytesType[1], DEC) + "." + String(data.BytesType[0], DEC));
			break;
		}

		case Register::REG_OPERATINGTIME1:
		{
			operatingTime[0] = data.BytesType[0];
			operatingTime[1] = data.BytesType[1];
			break;
		}

		case Register::REG_OPERATINGTIME2:
		{
			operatingTime[2] = data.BytesType[0];
			operatingTime[3] = data.BytesType[1];

			int totalSeconds = (operatingTime[3] << 24) | (operatingTime[2] << 16) | (operatingTime[1] << 8) | operatingTime[0];

			int mins = (int)(totalSeconds / 60);
			int secs = totalSeconds - (mins * 60);
			PrintMessage("Uptime: " + String(mins) + " minutes and " + String(secs) + " seconds.");

			break;
		}

		case Register::REPORT_DEVICE_ERROR_STATUS:
		{
			auto value = static_cast<HapticoreDeviceErrorStatus>(data.UnsignedType);
			bool error = value != HapticoreDeviceErrorStatus::NO_ERROR;
			if (error)
			{
				PrintMessage("Device Error Status: " + String(data.UnsignedType));
			}
			else
			{
				PrintMessage("Device Error Status: NO_ERROR.");
			}

			break;
		}

		case Register::REPORT_ENCODER_ANGLE:
		{
			PrintMessage("Angle: " + String(data.FloatType));
			break;
		}

		case Register::REPORT_ENCODER_MULTI_TURN_COUNT:
		{
			PrintMessage("Multi Turn Count: " + String(data.SignedType));
			break;
		}

		case Register::REPORT_TICK_INDEX:
		{
			PrintMessage("Tick Index: " + String(data.SignedType));
			break;
		}
	
		default:
		{
			break;
		}
	}
}
