#include "HapticoreLibrary.h"
#include "HapticoreConversionFactors.h"

namespace HapticoreLibrary
{
	static const byte PACKET_START_BYTE					= 0x26;
	static const byte PACKET_STOP_BYTE					= 0x0D;

	static const byte PACKET_LENGTH						= 6;
	static const byte PACKET_OFFSET_START_BYTE			= 0;
	static const byte PACKET_OFFSET_TYPE_BYTE			= 1;
	static const byte PACKET_OFFSET_DATA_HIGH_BYTE		= 2;
	static const byte PACKET_OFFSET_DATA_LOW_BYTE		= 3;
	static const byte PACKET_OFFSET_LRC_BYTE			= 4;
	static const byte PACKET_OFFSET_STOP_BYTE			= 5;

	static const int SERIAL_NO_SIZE						= 11;
	static const int HAPTICORE_ITEM_NO_SIZE				= 14;

	static const int  PACKET_TRANSMISSION_DELAY_MS		= 0;

	// Clear connection state disconnect event
	bool customParamSigned[7] 							= {0};
	unsigned short int customParamFactor[7] 			= {0};
	
	char manufacturerSerialNumber[SERIAL_NO_SIZE]		= {0};
	char hapticoreItemNumber[HAPTICORE_ITEM_NO_SIZE]	= {0};

	byte pushCalibrationActivateThreshold[4]			= {0};
	byte pushCalibrationDeactivateThreshold[4]			= {0};
	byte pullCalibrationActivateThreshold[4]			= {0};
	byte pullCalibrationDeactivateThreshold[4]			= {0};

	byte encoderCalibrationYOffset[4]					= {0};
	byte encoderCalibrationYAmplitude[4]				= {0};
	byte encoderCalibrationZOffset[4]					= {0};
	byte encoderCalibrationZAmplitude[4]				= {0};


	void HapticoreDevice::Connect(HardwareSerial* port, const long baudRate)
	{
		SerialPort = port;
		SerialPort->begin(baudRate, SERIAL_8N1);
		
		if (PrintMsgHandler)
		{
			PrintMsgHandler("Port opened successfully.");
		}
	}

	void HapticoreDevice::Disconnect()
	{
		SerialPort->end();
	}

	void HapticoreDevice::PollData()
	{
		if (SerialPort->available() > 0)
		{
			byte data = SerialPort->read();
			ParseReceivedByte(data);
		}
	}

	void HapticoreDevice::Request(const Register reg)
	{
		byte registerValue = static_cast<byte>(reg);

		if (reg == Register::REG_HAPTICORE_MANUFACTURER_SERIAL_NUMBER)
		{
			for (int i = 0; i < SERIAL_NO_SIZE; i++)
			{
				SerialWrite(Register::CMD_GET_REGISTER_VALUE, (byte)i, registerValue);
			}
		}
		else if (reg == Register::REG_HAPTICORE_ITEM_NUMBER)
		{
			// Request the first byte here, the rest is requested based on the response
			SerialWrite(Register::CMD_GET_REGISTER_VALUE, 0x0, registerValue);
		}
		else if (registerValue < 0x10)
		{
			SerialWrite(reg, 0x0, 0x0);
		}
		else
		{
			SerialWrite(Register::CMD_GET_REGISTER_VALUE, 0x0, registerValue);
		}
	}

	void HapticoreDevice::AttachDataReceivedHandler(const DataReceivedHandler_t handler)
	{
		DataReceivedHandler = handler;
	}

	void HapticoreDevice::AttachDebugLogHandler(const OutputMsgHandler_t handler)
	{
		PrintMsgHandler = handler;
	}

	void HapticoreDevice::DetachHandler()
	{
		DataReceivedHandler = nullptr;
		PrintMsgHandler = nullptr;
	}

	void HapticoreDevice::ClearState()
	{
		memset(customParamSigned, 0, sizeof(customParamSigned));
		memset(customParamFactor, 0, sizeof(customParamFactor));

		memset(manufacturerSerialNumber, '\0', sizeof manufacturerSerialNumber);
		memset(hapticoreItemNumber, '\0', sizeof hapticoreItemNumber);

		memset(pushCalibrationActivateThreshold, 0x0, sizeof pushCalibrationActivateThreshold);
		memset(pushCalibrationDeactivateThreshold, 0x0, sizeof pushCalibrationDeactivateThreshold);

		memset(pullCalibrationActivateThreshold, 0x0, sizeof pullCalibrationActivateThreshold);
		memset(pullCalibrationDeactivateThreshold, 0x0, sizeof pullCalibrationDeactivateThreshold);

		memset(encoderCalibrationYOffset, 0x0, sizeof encoderCalibrationYOffset);
		memset(encoderCalibrationYAmplitude, 0x0, sizeof encoderCalibrationYAmplitude);
		memset(encoderCalibrationZOffset, 0x0, sizeof encoderCalibrationZOffset);
		memset(encoderCalibrationZAmplitude, 0x0, sizeof encoderCalibrationZAmplitude);
	}

	void HapticoreDevice::NotifyDataReceivedSafe(const Register reg, const RegisterData data)
	{
		if (DataReceivedHandler)
		{
			DataReceivedHandler(reg, data);
		}
	}

	void HapticoreDevice::PrintMsgSafe(const String message)
	{
		if (PrintMsgHandler)
		{
			PrintMsgHandler(message);
		}
	}

	unsigned short int HapticoreDevice::ConvertDataToUshort(const byte* const packet)
	{
		unsigned short int payload = (packet[PACKET_OFFSET_DATA_HIGH_BYTE] << 8) + packet[PACKET_OFFSET_DATA_LOW_BYTE];
		return payload;
	}

	RegisterData HapticoreDevice::RequestBytesType(const byte* const packet)
	{
		RegisterData data;
		data.BytesType[1] = packet[PACKET_OFFSET_DATA_HIGH_BYTE];
		data.BytesType[0] = packet[PACKET_OFFSET_DATA_LOW_BYTE];

		return data;
	}

	RegisterData HapticoreDevice::RequestUshortFloatType(const byte* const packet, const unsigned short int conversionFactor)
	{
		RegisterData data;
		data.FloatType = 1.0f * ConvertDataToUshort(packet) / conversionFactor;

		return data;
	}

	RegisterData HapticoreDevice::RequestShortFloatType(const byte* const packet, const unsigned short int conversionFactor)
	{
		RegisterData data;
		data.FloatType = 1.0f * ((signed short int)ConvertDataToUshort(packet)) / conversionFactor;

		return data;
	}

	RegisterData HapticoreDevice::RequestSignedType(const byte* const packet)
	{
		RegisterData data;
		data.SignedType = (signed short int)ConvertDataToUshort(packet);
		
		return data;
	}
	
	RegisterData HapticoreDevice::RequestUnsignedType(const byte* const packet)
	{
		RegisterData data;
		data.UnsignedType = ConvertDataToUshort(packet);

		return data;
	}

	byte* HapticoreDevice::SendBytesType(const RegisterData data)
	{
		byte* dataBytes = new byte[2];
		dataBytes[0] = data.BytesType[0];
		dataBytes[1] = data.BytesType[1];
	}

	byte* HapticoreDevice::SendUnsignedType(const RegisterData data)
	{
		byte* dataBytes = new byte[2];
		dataBytes[0] = (byte)(data.UnsignedType);
		dataBytes[1] = (byte)(data.UnsignedType >> 8);

		return dataBytes;
	}

	byte* HapticoreDevice::SendUshortFloatType(const RegisterData data, unsigned short int conversionFactor)
	{
		RegisterData value;
		value.UnsignedType = data.FloatType * conversionFactor;
		
		return SendUnsignedType(value);
	}

	void HapticoreDevice::ParseReceivedByte(const byte data)
	{
		parseBufferIdx = (parseBufferIdx + 1) % PACKET_LENGTH;
		parseBuffer[parseBufferIdx] = data;

		// Check if the present data is valid throuch CRC
		// Start validating at the next index!
		bool startCorrect = parseBuffer[(parseBufferIdx + PACKET_OFFSET_START_BYTE + 1) % PACKET_LENGTH] == PACKET_START_BYTE;
		bool stopCorrect = parseBuffer[(parseBufferIdx + PACKET_OFFSET_STOP_BYTE + 1) % PACKET_LENGTH] == PACKET_STOP_BYTE;
		bool lrcCorrect = parseBuffer[(parseBufferIdx + PACKET_OFFSET_LRC_BYTE + 1) % PACKET_LENGTH] == 
			(parseBuffer[(parseBufferIdx + PACKET_OFFSET_TYPE_BYTE + 1) % PACKET_LENGTH] ^
			parseBuffer[(parseBufferIdx + PACKET_OFFSET_DATA_HIGH_BYTE + 1) % PACKET_LENGTH] ^
			parseBuffer[(parseBufferIdx + PACKET_OFFSET_DATA_LOW_BYTE + 1) % PACKET_LENGTH]);


		if (startCorrect && stopCorrect && lrcCorrect)
		{
			byte alignedBuffer[6];
			alignedBuffer[0] = parseBuffer[(parseBufferIdx + PACKET_OFFSET_START_BYTE + 1) % PACKET_LENGTH];
			alignedBuffer[1] = parseBuffer[(parseBufferIdx + PACKET_OFFSET_TYPE_BYTE + 1) % PACKET_LENGTH];
			alignedBuffer[2] = parseBuffer[(parseBufferIdx + PACKET_OFFSET_DATA_HIGH_BYTE + 1) % PACKET_LENGTH];
			alignedBuffer[3] = parseBuffer[(parseBufferIdx + PACKET_OFFSET_DATA_LOW_BYTE + 1) % PACKET_LENGTH];
			alignedBuffer[4] = parseBuffer[(parseBufferIdx + PACKET_OFFSET_LRC_BYTE + 1) % PACKET_LENGTH];
			alignedBuffer[5] = parseBuffer[(parseBufferIdx + PACKET_OFFSET_STOP_BYTE + 1) % PACKET_LENGTH];

			HandlePacket(alignedBuffer);
		}
	}

	void HapticoreDevice::SerialWrite(const Register reg, const byte dataHighByte, const byte dataLowByte)
	{
		auto packet = new byte[PACKET_LENGTH];
		packet[PACKET_OFFSET_START_BYTE] = PACKET_START_BYTE;
		packet[PACKET_OFFSET_TYPE_BYTE] = static_cast<byte>(reg);
		packet[PACKET_OFFSET_DATA_HIGH_BYTE] = dataHighByte;
		packet[PACKET_OFFSET_DATA_LOW_BYTE] = dataLowByte;
		packet[PACKET_OFFSET_LRC_BYTE] = packet[PACKET_OFFSET_TYPE_BYTE] 
			^ packet[PACKET_OFFSET_DATA_HIGH_BYTE] 
			^ packet[PACKET_OFFSET_DATA_LOW_BYTE];

		packet[PACKET_OFFSET_STOP_BYTE] = PACKET_STOP_BYTE;

		SerialPort->write(packet, PACKET_LENGTH);
		delay(PACKET_TRANSMISSION_DELAY_MS);
	}

	void HapticoreDevice::HandlePacket(const byte* const packet)
	{
		Register reg = static_cast<Register>(packet[PACKET_OFFSET_TYPE_BYTE]);
		RegisterData data;

		switch (reg)
		{
			// Readonly
			case Register::REG_CONTROLLER_ID:
			case Register::REG_FW_VERSION:
			case Register::REG_COMMUNICATION_PROTOCOL_VERSION:
			case Register::REG_HAPTICORE_LIBRARY_VERSION:
			case Register::REG_CONTROLLER_HARDWARE_REVISION:
			{
				data = RequestUnsignedType(packet);
				break;
			}

			case Register::REG_HAPTICORE_MANUFACTURER_SERIAL_NUMBER:
			{
				byte index = packet[PACKET_OFFSET_DATA_HIGH_BYTE];
				char value = (char)packet[PACKET_OFFSET_DATA_LOW_BYTE];

				manufacturerSerialNumber[index] = value;
				if (value == '\0')
				{
					// Notify when complete
					data.StringType = manufacturerSerialNumber;
					break;
				}
				else
				{
					// Don't forward while not complete
					return;
				}
			}

			case Register::REG_HAPTICORE_TYPE:
			{
				data = RequestBytesType(packet);
				break;
			}

			case Register::REG_HAPTICORE_PROTOTYPE_VERSION:
			case Register::REG_HAPTICORE_PROTOTYPE_NUMBER:
			{
				data = RequestUnsignedType(packet);
				break;
			}

			case Register::REG_HAPTICORE_PROTOTYPE_LOT:
			case Register::REG_HAPTICORE_DIAMETER:
			{
				data = RequestBytesType(packet);
				break;
			}

			case Register::REG_HAPTICORE_PRODUCT_NUMBER:
			{
				data = RequestUnsignedType(packet);
				break;
			}

			case Register::REG_HAPTICORE_VARIANT_NUMBER:
			{
				data = RequestBytesType(packet);
				break;
			}

			case Register::REG_HAPTICORE_ITEM_NUMBER:
			{
				byte index = packet[PACKET_OFFSET_DATA_HIGH_BYTE];
				char value = (char)packet[PACKET_OFFSET_DATA_LOW_BYTE];

				hapticoreItemNumber[index] = value;
				if (index == HAPTICORE_ITEM_NO_SIZE - 1 || value == '\0')
				{
					// Notify when complete
					data.StringType = hapticoreItemNumber;
					break;
				}
				else
				{
					// Request next packet
					SerialWrite(Register::CMD_GET_REGISTER_VALUE, index + 1, static_cast<byte>(Register::REG_HAPTICORE_ITEM_NUMBER));

					// Don't forward while not complete
					return;
				}
			}

			case Register::REG_OPERATINGTIME1:
			case Register::REG_OPERATINGTIME2:
			{
				data = RequestBytesType(packet);
				break;
			}

			case Register::REG_ENCODER_TEMPERATURE:
			{
				data = RequestShortFloatType(packet, FACTOR_TEMPERATURE);
				break;
			}

			case Register::REG_SYSTEM_CONFIGURATION:
			{
				data = RequestBytesType(packet);
				break;
			}

			case Register::REG_REPORT_TYPE:
			case Register::REG_REPORT_FLAGS:
			case Register::REG_REPORT_FREQUENCY:
			{
				data = RequestUnsignedType(packet);
				break;
			}

			case Register::REG_LED_RGB_RED:
			case Register::REG_LED_RGB_GREEN:
			case Register::REG_LED_RGB_BLUE:
			{
				data = RequestBytesType(packet);
				break;
			}

			case Register::REG_IDLE_DETECTION_VELOCITY_THRESHOLD:
			{
				data = RequestUshortFloatType(packet, FACTOR_VELOCITY_THRESHOLD);
				break;
			}

			case Register::REG_IDLE_DETECTION_TIMEOUT:
			{
				data = RequestUshortFloatType(packet, FACTOR_TIMEOUT);
				break;
			}

			case Register::REG_IDLE_DETECTION_CURRENT_RAMP_DOWN_SLOPE:
			{
				data = RequestShortFloatType(packet, FACTOR_RAMP_DOWN_SLOPE);
				break;
			}
			
			case Register::REG_CURRENT_CONTROLLER_KP:
			{
				data = RequestUshortFloatType(packet, FACTOR_KP);
				break;
			}
			
			case Register::REG_CURRENT_CONTROLLER_IDLE_CURRENT:
			{
				data = RequestShortFloatType(packet, FACTOR_CURRENT);
				break;
			}

			case Register::REG_CURRENT_CONTROLLER_MODE:
			{
				data = RequestUnsignedType(packet);
				break;
			}

			case Register::REG_CURRENT_CONTROLLER_KI:
			{
				data = RequestUshortFloatType(packet, FACTOR_KI);
				break;
			}

			case Register::REG_CURRENT_CONTROLLER_COIL_RESISTANCE:
			{
				data = RequestUshortFloatType(packet, FACTOR_RESISTANCE);
				break;
			}

			case Register::REG_CURRENT_CONTROLLER_COIL_INDUCTANCE:
			{
				data = RequestUshortFloatType(packet, FACTOR_INDUCTANCE);
				break;
			}

			case Register::REG_ENCODER_ANGLE_FILTER_R:
			{
				data = RequestUshortFloatType(packet, FACTOR_R_ANGLE);
				break;
			}

			case Register::REG_ENCODER_VELOCITY_FILTER_R:
			{
				data = RequestUshortFloatType(packet, FACTOR_R_VELOCITY);
				break;
			}

			case Register::REG_CURRENT_CONTROLLER_SUPPLY_VOLTAGE_OVERRIDE:
			{
				data = RequestUnsignedType(packet);
				break;
			}
			
			case Register::REG_CURRENT_CONTROLLER_SUPPLY_VOLTAGE:
			{
				data = RequestUshortFloatType(packet, FACTOR_VOLTAGE);
				break;
			}

			case Register::REG_COIL_DRIVER_PWM_POLARITY:
			case Register::REG_COIL_DRIVER_STOP_MODE:
			case Register::REG_CURRENT_SENSE_POLARITY:
			{
				data = RequestBytesType(packet);
				break;
			}
			
			case Register::REG_ENCODER_MODE:
			{
				data = RequestUnsignedType(packet);
				break;
			}
			
			case Register::REG_ENCODER_ANGLE:
			{
				data = RequestUnsignedType(packet);
				break;
			}
			
			case Register::REG_ENCODER_TIME_TRIGGERED_ANGLE_INCREMENT:
			{
				data = RequestUshortFloatType(packet, FACTOR_TIME_TRIGGERED_ANGLE_INCREMENT);
				break;
			}
			
			case Register::REG_ENCODER_VELOCITY_FILTER_Q:
			{
				data = RequestUshortFloatType(packet, FACTOR_Q_VELOCITY);
				break;
			}
			
			case Register::REG_ENCODER_VELOCITY_NOISE_THRESHOLD:
			{
				data = RequestUshortFloatType(packet, FACTOR_VELOCITY_THRESHOLD);
				break;
			}
			
			case Register::REG_ENCODER_DIRECTION:
			{
				data = RequestUnsignedType(packet);
				break;
			}
			
			case Register::REG_ENCODER_ANGLE_FILTER_Q:
			{
				data = RequestUshortFloatType(packet, FACTOR_Q_ANGLE);
				break;
			}
			
			case Register::REG_ENCODER_ANGLE_RESOLUTION:
			{
				data = RequestUshortFloatType(packet, FACTOR_ENCODER_ANGLE_RESOLUTION);
				break;
			}
			
			case Register::REG_ENCODER_CALIBRATION_Y_OFFSET:
			{
				byte index = packet[PACKET_OFFSET_DATA_HIGH_BYTE];
				byte value = packet[PACKET_OFFSET_DATA_LOW_BYTE];

				encoderCalibrationYOffset[index] = value;
				if (index == 0x03)
				{
					data.FloatType = ((float*)encoderCalibrationYOffset)[0];
					// Notify when complete
					break;
				}
				else
				{
					// Don't forward while not complete
					return;
				}
			}
			
			case Register::REG_ENCODER_CALIBRATION_Y_AMPLITUDE:
			{
				byte index = packet[PACKET_OFFSET_DATA_HIGH_BYTE];
				byte value = packet[PACKET_OFFSET_DATA_LOW_BYTE];

				encoderCalibrationYAmplitude[index] = value;
				if (index == 0x03)
				{
					data.FloatType = ((float*)encoderCalibrationYAmplitude)[0];
					// Notify when complete
					break;
				}
				else
				{
					// Don't forward while not complete
					return;
				}
			}
			
			case Register::REG_ENCODER_CALIBRATION_Z_OFFSET:
			{
				byte index = packet[PACKET_OFFSET_DATA_HIGH_BYTE];
				byte value = packet[PACKET_OFFSET_DATA_LOW_BYTE];

				encoderCalibrationZOffset[index] = value;
				if (index == 0x03)
				{
					data.FloatType = ((float*)encoderCalibrationZOffset)[0];
					// Notify when complete
					break;
				}
				else
				{
					// Don't forward while not complete
					return;
				}
			}
			
			case Register::REG_ENCODER_CALIBRATION_Z_AMPLITUDE:
			{
				byte index = packet[PACKET_OFFSET_DATA_HIGH_BYTE];
				byte value = packet[PACKET_OFFSET_DATA_LOW_BYTE];

				encoderCalibrationZAmplitude[index] = value;
				if (index == 0x03)
				{
					data.FloatType = ((float*)encoderCalibrationZAmplitude)[0];
					// Notify when complete
					break;
				}
				else
				{
					// Don't forward while not complete
					return;
				}
			}

			case Register::REG_ENCODER_OFFSET:
			{
				data = RequestShortFloatType(packet, FACTOR_ANGLE);
				break;
			}

			case Register::REG_ENCODER_SCALING:
			{
				data = RequestUshortFloatType(packet, FACTOR_SCALING);
				break;
			}
			
			case Register::REG_DEGAUSS_MODE:
			case Register::REG_DEGAUSS_FREQUENCY:
			{
				data = RequestUnsignedType(packet);
				break;
			}
			
			case Register::REG_DEGAUSS_CURRENT:
			{
				data = RequestUshortFloatType(packet, FACTOR_CURRENT);
				break;
			}
			
			case Register::REG_DEGAUSS_DURATION:
			{
				data = RequestUshortFloatType(packet, FACTOR_DURATION);
				break;
			}
			
			case Register::REG_PUSH_ACTIVATE_THRESHOLD:
			{
				byte index = packet[PACKET_OFFSET_DATA_HIGH_BYTE];
				byte value = packet[PACKET_OFFSET_DATA_LOW_BYTE];

				pushCalibrationActivateThreshold[index] = value;
				if (index == 0x03)
				{
					data.FloatType = ((float*)pushCalibrationActivateThreshold)[0];
					// Notify when complete
					break;
				}
				else
				{
					// Don't forward while not complete
					return;
				}
			}
			
			case Register::REG_PUSH_DEACTIVATE_THRESHOLD:
			{
				byte index = packet[PACKET_OFFSET_DATA_HIGH_BYTE];
				byte value = packet[PACKET_OFFSET_DATA_LOW_BYTE];

				pushCalibrationDeactivateThreshold[index] = value;
				if (index == 0x03)
				{
					data.FloatType = ((float*)pushCalibrationDeactivateThreshold)[0];
					// Notify when complete
					break;
				}
				else
				{
					// Don't forward while not complete
					return;
				}
			}
			
			case Register::REG_PULL_ACTIVATE_THRESHOLD:
			{
				byte index = packet[PACKET_OFFSET_DATA_HIGH_BYTE];
				byte value = packet[PACKET_OFFSET_DATA_LOW_BYTE];

				pullCalibrationActivateThreshold[index] = value;
				if (index == 0x03)
				{
					data.FloatType = ((float*)pullCalibrationActivateThreshold)[0];
					// Notify when complete
					break;
				}
				else
				{
					// Don't forward while not complete
					return;
				}
			}
			
			case Register::REG_PULL_DEACTIVATE_THRESHOLD:
			{
				byte index = packet[PACKET_OFFSET_DATA_HIGH_BYTE];
				byte value = packet[PACKET_OFFSET_DATA_LOW_BYTE];

				pullCalibrationDeactivateThreshold[index] = value;
				if (index == 0x03)
				{
					data.FloatType = ((float*)pullCalibrationDeactivateThreshold)[0];
					// Notify when complete
					break;
				}
				else
				{
					// Don't forward while not complete
					return;
				}
			}
			
			// Haptic Function Tick
			case Register::REG_TICK_INDEX_COUNT_MODE:
			{
				data = RequestBytesType(packet);
				break;
			}

			case Register::REG_TICK_DEGAUSS_CURRENT:
			{
				data = RequestShortFloatType(packet, FACTOR_CURRENT);
				break;
			}
			
			case Register::REG_TICK_DEGAUSS_DURATION:
			{
				data = RequestUshortFloatType(packet, FACTOR_DURATION);
				break;
			}
			
			case Register::REG_TICK_EXPONENTIAL_RAMP_DOWN_VELOCITY_THRESHOLD:
			{
				data = RequestUshortFloatType(packet, FACTOR_VELOCITY_THRESHOLD);
				break;
			}
			
			case Register::REG_TICK_EXPONENTIAL_RAMP_DOWN_FACTOR:
			{
				data = RequestUshortFloatType(packet, FACTOR_RAMP_DOWN);
				break;
			}
			
			case Register::REG_TICK_EXPONENTIAL_RAMP_DOWN_CURRENT_MIN_PERCENT:
			{
				data = RequestUshortFloatType(packet, FACTOR_PERCENT);
				break;
			}
			
			case Register::REG_TICK_MODE:
			case Register::REG_TICK_ENABLE:
			{
				data = RequestUnsignedType(packet);
				break;
			}
			
			case Register::REG_TICK_CURRENT:
			{
				data = RequestShortFloatType(packet, FACTOR_CURRENT);
				break;
			}
			
			case Register::REG_TICK_ANGLE_CW:
			case Register::REG_TICK_ANGLE_CCW:
			{
				data = RequestUshortFloatType(packet, FACTOR_ANGLE);
				break;
			}
			
			case Register::REG_TICK_DURATION_MIN:
			case Register::REG_TICK_DURATION_MAX:
			{
				data = RequestUshortFloatType(packet, FACTOR_DURATION);
				break;
			}

			case Register::REG_TICK_VELOCITY_FACTOR:
			{
				data = RequestShortFloatType(packet, FACTOR_VELOCITY_FACTOR);
				break;
			}

			case Register::REG_TICK_INDEX_WINDOW:
			{
				data = RequestUshortFloatType(packet, FACTOR_WINDOW);
				break;
			}

			case Register::REG_TICK_STICKINESS_PREVENTION_FACTOR:
			{
				data = RequestUshortFloatType(packet, FACTOR_STICKINESS);
				break;
			}

			case Register::REG_TICK_ACTIVE_DIRECTION:
			{
				data = RequestUnsignedType(packet);
				break;
			}

			case Register::REG_TICK_FREEWHEELING_VELOCITY_THRESHOLD:
			{
				data = RequestUshortFloatType(packet, FACTOR_VELOCITY_THRESHOLD);
				break;
			}

			case Register::REG_TICK_FREEWHEELING_EXTENSION_TIME:
			{
				data = RequestUshortFloatType(packet, FACTOR_DURATION);
				break;
			}

			case Register::REG_TICK_WINDOW:
			{
				data = RequestUshortFloatType(packet, FACTOR_WINDOW);
				break;
			}

			case Register::REG_TICK_START_ANGLE:
			case Register::REG_TICK_STOP_ANGLE:
			{
				data = RequestShortFloatType(packet, FACTOR_ANGLE);
				break;
			}

			// Haptic Function Barrier
			case Register::REG_BARRIER_ENABLE:
			{
				data = RequestUnsignedType(packet);
				break;
			}

			case Register::REG_BARRIER_CURRENT:
			{
				data = RequestShortFloatType(packet, FACTOR_CURRENT);
				break;
			}

			case Register::REG_BARRIER_START_ANGLE:
			case Register::REG_BARRIER_STOP_ANGLE:
			{
				data = RequestShortFloatType(packet, FACTOR_ANGLE);
				break;
			}

			case Register::REG_BARRIER_POLARITY:
			case Register::REG_BARRIER_LIMIT_ANGLE:
			{
				data = RequestUnsignedType(packet);
				break;
			}

			// Haptic Function Current
			case Register::REG_CURRENT_ENABLE:
			{
				data = RequestUnsignedType(packet);
				break;
			}

			case Register::REG_CURRENT_START_ANGLE:
			case Register::REG_CURRENT_STOP_ANGLE:
			{
				data = RequestShortFloatType(packet, FACTOR_ANGLE);
				break;
			}

			case Register::REG_CURRENT_START_CURRENT_CW:
			case Register::REG_CURRENT_STOP_CURRENT_CW:
			case Register::REG_CURRENT_START_CURRENT_CCW:
			case Register::REG_CURRENT_STOP_CURRENT_CCW:
			{
				data = RequestShortFloatType(packet, FACTOR_CURRENT);
				break;
			}

			case Register::REG_CURRENT_FREEWHEELING_VELOCITY_THRESHOLD:
			{
				data = RequestUshortFloatType(packet, FACTOR_VELOCITY_THRESHOLD);
				break;
			}

			case Register::REG_CURRENT_FREEWHEELING_EXTENSION_TIME:
			{
				data = RequestUshortFloatType(packet, FACTOR_DURATION);
				break;
			}

			case Register::REG_CURRENT_ACTIVE_DIRECTION:
			{
				data = RequestUnsignedType(packet);
				break;
			}

			case Register::REG_CURRENT_STREAMING_ENABLE:
			{
				data = RequestBytesType(packet);
				break;
			}

			case Register::REG_CURRENT_STREAMING_CURRENT:
			{
				data = RequestShortFloatType(packet, FACTOR_CURRENT);
				break;
			}

			// Haptic Function Torque
			case Register::REG_TORQUE_ENABLE:
			{
				data = RequestUnsignedType(packet);
				break;
			}

			case Register::REG_TORQUE_START_ANGLE:
			case Register::REG_TORQUE_STOP_ANGLE:
			{
				data = RequestShortFloatType(packet, FACTOR_ANGLE);
				break;
			}

			case Register::REG_TORQUE_START_FACTOR_CW:
			case Register::REG_TORQUE_STOP_FACTOR_CW:
			case Register::REG_TORQUE_START_FACTOR_CCW:
			case Register::REG_TORQUE_STOP_FACTOR_CCW:
			{
				data = RequestShortFloatType(packet, FACTOR_CURRENT);
				break;
			}

			case Register::REG_TORQUE_FREEWHEELING_VELOCITY_THRESHOLD:
			{
				data = RequestUshortFloatType(packet, FACTOR_VELOCITY_THRESHOLD);
				break;
			}

			case Register::REG_TORQUE_FREEWHEELING_EXTENSION_TIME:
			{
				data = RequestUshortFloatType(packet, FACTOR_DURATION);
				break;
			}

			case Register::REG_TORQUE_ACTIVE_DIRECTION:
			{
				data = RequestUnsignedType(packet);
				break;
			}

			// Haptic Function Lock
			case Register::REG_LOCK_ENABLE:
			case Register::REG_LOCK_DIRECTION:
			{
				data = RequestUnsignedType(packet);
				break;
			}

			case Register::REG_LOCK_CURRENT:
			{
				data = RequestShortFloatType(packet, FACTOR_CURRENT);
				break;
			}

			// Haptic Function Single Tick
			case Register::REG_SINGLE_TICK_ENABLE:
			case Register::REG_SINGLE_TICK_ACTIVE_DIRECTION:
			case Register::REG_SINGLE_TICK_MODE:
			{
				data = RequestUnsignedType(packet);
				break;
			}

			case Register::REG_SINGLE_TICK_ANGLE:
			{
				data = RequestUshortFloatType(packet, FACTOR_ANGLE);
				break;
			}
			
			case Register::REG_SINGLE_TICK_CURRENT:
			{
				data = RequestShortFloatType(packet, FACTOR_CURRENT);
				break;
			}

			case Register::REG_SINGLE_TICK_DURATION_MIN:
			case Register::REG_SINGLE_TICK_DURATION_MAX:
			{
				data = RequestUshortFloatType(packet, FACTOR_DURATION);
				break;
			}

			case Register::REG_SINGLE_TICK_WINDOW:
			{
				data = RequestUshortFloatType(packet, FACTOR_WINDOW);
				break;
			}

			case Register::REG_SINGLE_TICK_VELOCITY_FACTOR:
			{
				data = RequestShortFloatType(packet, FACTOR_VELOCITY_FACTOR);
				break;
			}

			// Haptic Function Haptics Generator
			case Register::REG_HAPTICS_GENERATOR_ENABLE:
			{
				data = RequestUnsignedType(packet);
				break;
			}

			// Custom Mode
			case Register::REG_CUSTOM_MODE_ENABLE:
			case Register::REG_CUSTOM_MODE:
			{
				data = RequestUnsignedType(packet);
				break;
			}

			case Register::REG_CUSTOM_PARAMETER1:
			{
				if (customParamSigned[0])
				{
					data = RequestShortFloatType(packet, customParamFactor[0]);
				}
				else
				{
					data = RequestUshortFloatType(packet, customParamFactor[0]);
				}
				break;
			}

			case Register::REG_CUSTOM_PARAMETER2:
			{
				if (customParamSigned[1])
				{
					data = RequestShortFloatType(packet, customParamFactor[1]);
				}
				else
				{
					data = RequestUshortFloatType(packet, customParamFactor[1]);
				}
				break;
			}

			case Register::REG_CUSTOM_PARAMETER3:
			{
				if (customParamSigned[2])
				{
					data = RequestShortFloatType(packet, customParamFactor[2]);
				}
				else
				{
					data = RequestUshortFloatType(packet, customParamFactor[2]);
				}
				break;
			}

			case Register::REG_CUSTOM_PARAMETER4:
			{
				if (customParamSigned[3])
				{
					data = RequestShortFloatType(packet, customParamFactor[3]);
				}
				else
				{
					data = RequestUshortFloatType(packet, customParamFactor[3]);
				}
				break;
			}

			case Register::REG_CUSTOM_PARAMETER5:
			{
				if (customParamSigned[4])
				{
					data = RequestShortFloatType(packet, customParamFactor[4]);
				}
				else
				{
					data = RequestUshortFloatType(packet, customParamFactor[4]);
				}
				break;
			}

			case Register::REG_CUSTOM_PARAMETER6:
			{
				if (customParamSigned[5])
				{
					data = RequestShortFloatType(packet, customParamFactor[5]);
				}
				else
				{
					data = RequestUshortFloatType(packet, customParamFactor[5]);
				}
				break;
			}

			case Register::REG_CUSTOM_PARAMETER7:
			{
				if (customParamSigned[6])
				{
					data = RequestShortFloatType(packet, customParamFactor[6]);
				}
				else
				{
					data = RequestUshortFloatType(packet, customParamFactor[6]);
				}
				break;
			}

			case Register::REG_CUSTOM_PARAMETER1_PROPERTIES:
			{
				data = RequestUnsignedType(packet);
				customParamSigned[0] = (data.UnsignedType >> 15);
				customParamFactor[0] = (data.UnsignedType & 0x7FFF);
				return;
			}

			case Register::REG_CUSTOM_PARAMETER2_PROPERTIES:
			{
				data = RequestUnsignedType(packet);
				customParamSigned[1] = (data.UnsignedType >> 15);
				customParamFactor[1] = (data.UnsignedType & 0x7FFF);
				return;
			}

			case Register::REG_CUSTOM_PARAMETER3_PROPERTIES:
			{
				data = RequestUnsignedType(packet);
				customParamSigned[2] = (data.UnsignedType >> 15);
				customParamFactor[2] = (data.UnsignedType & 0x7FFF);
				return;
			}

			case Register::REG_CUSTOM_PARAMETER4_PROPERTIES:
			{
				data = RequestUnsignedType(packet);
				customParamSigned[3] = (data.UnsignedType >> 15);
				customParamFactor[3] = (data.UnsignedType & 0x7FFF);
				return;
			}

			case Register::REG_CUSTOM_PARAMETER5_PROPERTIES:
			{
				data = RequestUnsignedType(packet);
				customParamSigned[4] = (data.UnsignedType >> 15);
				customParamFactor[4] = (data.UnsignedType & 0x7FFF);
				return;
			}

			case Register::REG_CUSTOM_PARAMETER6_PROPERTIES:
			{
				data = RequestUnsignedType(packet);
				customParamSigned[5] = (data.UnsignedType >> 15);
				customParamFactor[5] = (data.UnsignedType & 0x7FFF);
				return;
			}

			case Register::REG_CUSTOM_PARAMETER7_PROPERTIES:
			{
				data = RequestUnsignedType(packet);
				customParamSigned[6] = (data.UnsignedType >> 15);
				customParamFactor[6] = (data.UnsignedType & 0x7FFF);
				return;
			}

			// Reports
			case Register::REPORT_ENCODER_ANGLE:
			{
				data = RequestUshortFloatType(packet, FACTOR_ENCODER_ANGLE_READ);
				break;
			}

			case Register::REPORT_ENCODER_VELOCITY:
			{
				data = RequestShortFloatType(packet, FACTOR_ENCODER_VELOCITY);
				break;
			}

			case Register::REPORT_PUSH_PULL_VALUE:
			{
				data = RequestShortFloatType(packet, FACTOR_PUSH_PULL_VALUE);
				break;
			}

			case Register::REPORT_TICK_INDEX:
			{
				data = RequestSignedType(packet);
				break;
			}

			case Register::REPORT_PUSH_PULL_STATE:
			{
				data = RequestUnsignedType(packet);
				break;
			}

			case Register::REPORT_ENCODER_MULTI_TURN_COUNT:
			{
				data = RequestSignedType(packet);
				break;
			}

			case Register::REPORT_ENCODER_TEMPERATURE:
			{
				data = RequestShortFloatType(packet, FACTOR_TEMPERATURE);
				break;
			}

			case Register::REPORT_COIL_CURRENT:
			{
				data = RequestShortFloatType(packet, FACTOR_CURRENT);
				break;
			}

			case Register::REPORT_DEVICE_ERROR_STATUS:
			case Register::REPORT_ENCODER_CALIBRATION_STATUS:
			case Register::REPORT_PUSH_CALIBRATION_STATUS:
			case Register::REPORT_PULL_CALIBRATION_STATUS:
			case Register::REPORT_DEGAUSS_STATUS:
			{
				data = RequestUnsignedType(packet);
				break;
			}

			case Register::REPORT_DEVICE_CONNECTION_STATE:
			{
				data = RequestUnsignedType(packet);
				bool disconnected = static_cast<unsigned short int>(data.UnsignedType) 
					== static_cast<unsigned short int>(HapticoreDeviceConnectionState::DISCONNECTED);
				
				if (disconnected)
				{
					// Clear the temporary state
					ClearState();
					PrintMsgSafe("Encoder disconnected.");
					break;
				}
			}

			case Register::REPORT_SUPPLY_VOLTAGE:
			{
				data = RequestUshortFloatType(packet, FACTOR_VOLTAGE);
				break;
			}

			default:
			{
				return;
			}
		}

		NotifyDataReceivedSafe(reg, data);
	}

	void HapticoreDevice::Send(const Register reg, const RegisterData data)
	{
		byte registerValue = static_cast<byte>(reg);
		byte* packetData = nullptr;

		if (registerValue >= 0x29)
		{
			switch (reg)
			{
				case Register::REG_SYSTEM_CONFIGURATION:
				{
					packetData = SendBytesType(data);
					break;
				}

				case Register::REG_REPORT_TYPE:
				case Register::REG_REPORT_FLAGS:
				case Register::REG_REPORT_FREQUENCY:
				{
					packetData = SendUnsignedType(data);
					break;
				}

				case Register::REG_LED_RGB_RED:
				case Register::REG_LED_RGB_GREEN:
				case Register::REG_LED_RGB_BLUE:
				{
					packetData = SendBytesType(data);
					break;
				}

				case Register::REG_IDLE_DETECTION_VELOCITY_THRESHOLD:
				{
					packetData = SendUshortFloatType(data, FACTOR_VELOCITY_THRESHOLD);
					break;
				}

				case Register::REG_IDLE_DETECTION_TIMEOUT:
				{
					packetData = SendUshortFloatType(data, FACTOR_TIMEOUT);
					break;
				}

				case Register::REG_IDLE_DETECTION_CURRENT_RAMP_DOWN_SLOPE:
				{
					packetData = SendUshortFloatType(data, FACTOR_RAMP_DOWN_SLOPE);
					break;
				}
				
				case Register::REG_CURRENT_CONTROLLER_KP:
				{
					packetData = SendUshortFloatType(data, FACTOR_KP);
					break;
				}
				
				case Register::REG_CURRENT_CONTROLLER_IDLE_CURRENT:
				{
					packetData = SendUshortFloatType(data, FACTOR_CURRENT);
					break;
				}

				case Register::REG_CURRENT_CONTROLLER_MODE:
				{
					packetData = SendUnsignedType(data);
					break;
				}

				case Register::REG_CURRENT_CONTROLLER_KI:
				{
					packetData = SendUshortFloatType(data, FACTOR_KI);
					break;
				}

				case Register::REG_CURRENT_CONTROLLER_COIL_RESISTANCE:
				{
					packetData = SendUshortFloatType(data, FACTOR_KI);
					break;
				}

				case Register::REG_CURRENT_CONTROLLER_COIL_INDUCTANCE:
				{
					packetData = SendUshortFloatType(data, FACTOR_INDUCTANCE);
					break;
				}

				case Register::REG_ENCODER_ANGLE_FILTER_R:
				{
					packetData = SendUshortFloatType(data, FACTOR_R_ANGLE);
					break;
				}

				case Register::REG_ENCODER_VELOCITY_FILTER_R:
				{
					packetData = SendUshortFloatType(data, FACTOR_R_VELOCITY);
					break;
				}

				case Register::REG_CURRENT_CONTROLLER_SUPPLY_VOLTAGE_OVERRIDE:
				{
					packetData = SendUnsignedType(data);
					break;
				}
				
				case Register::REG_CURRENT_CONTROLLER_SUPPLY_VOLTAGE:
				{
					packetData = SendUshortFloatType(data, FACTOR_VOLTAGE);
					break;
				}

				case Register::REG_COIL_DRIVER_PWM_POLARITY:
				case Register::REG_COIL_DRIVER_STOP_MODE:
				case Register::REG_CURRENT_SENSE_POLARITY:
				{
					packetData = SendBytesType(data);
					break;
				}

				case Register::REG_ENCODER_MODE:
				{
					packetData = SendUnsignedType(data);
					break;
				}
				
				case Register::REG_ENCODER_ANGLE:
				{
					packetData = SendUshortFloatType(data, FACTOR_ANGLE);
					break;
				}
				
				case Register::REG_ENCODER_TIME_TRIGGERED_ANGLE_INCREMENT:
				{
					packetData = SendUshortFloatType(data, FACTOR_TIME_TRIGGERED_ANGLE_INCREMENT);
					break;
				}
				
				case Register::REG_ENCODER_VELOCITY_FILTER_Q:
				{
					packetData = SendUshortFloatType(data, FACTOR_Q_VELOCITY);
					break;
				}
				
				case Register::REG_ENCODER_VELOCITY_NOISE_THRESHOLD:
				{
					packetData = SendUshortFloatType(data, FACTOR_VELOCITY_THRESHOLD);
					break;
				}
				
				case Register::REG_ENCODER_DIRECTION:
				{
					packetData = SendUnsignedType(data);
					break;
				}
				
				case Register::REG_ENCODER_ANGLE_FILTER_Q:
				{
					packetData = SendUshortFloatType(data, FACTOR_Q_ANGLE);
					break;
				}
				
				case Register::REG_ENCODER_ANGLE_RESOLUTION:
				{
					packetData = SendUshortFloatType(data, FACTOR_ENCODER_ANGLE_RESOLUTION);
					break;
				}

				case Register::REG_ENCODER_OFFSET:
				{
					packetData = SendUshortFloatType(data, FACTOR_ANGLE);
					break;
				}

				case Register::REG_ENCODER_SCALING:
				{
					packetData = SendUshortFloatType(data, FACTOR_SCALING);
					break;
				}
				
				case Register::REG_DEGAUSS_MODE:
				case Register::REG_DEGAUSS_FREQUENCY:
				{
					packetData = SendUnsignedType(data);
					break;
				}
				
				case Register::REG_DEGAUSS_CURRENT:
				{
					packetData = SendUshortFloatType(data, FACTOR_CURRENT);
					break;
				}
				
				case Register::REG_DEGAUSS_DURATION:
				{
					packetData = SendUshortFloatType(data, FACTOR_DURATION);
					break;
				}
				
				// Haptic Function Tick
				case Register::REG_TICK_INDEX_COUNT_MODE:
				{
					packetData = SendBytesType(data);
					break;
				}

				case Register::REG_TICK_DEGAUSS_CURRENT:
				{
					packetData = SendUshortFloatType(data, FACTOR_CURRENT);
					break;
				}
				
				case Register::REG_TICK_DEGAUSS_DURATION:
				{
					packetData = SendUshortFloatType(data, FACTOR_DURATION);
					break;
				}
				
				case Register::REG_TICK_EXPONENTIAL_RAMP_DOWN_VELOCITY_THRESHOLD:
				{
					packetData = SendUshortFloatType(data, FACTOR_VELOCITY_THRESHOLD);
					break;
				}
				
				case Register::REG_TICK_EXPONENTIAL_RAMP_DOWN_FACTOR:
				{
					packetData = SendUshortFloatType(data, FACTOR_RAMP_DOWN);
					break;
				}
				
				case Register::REG_TICK_EXPONENTIAL_RAMP_DOWN_CURRENT_MIN_PERCENT:
				{
					packetData = SendUshortFloatType(data, FACTOR_PERCENT);
					break;
				}
				
				case Register::REG_TICK_MODE:
				case Register::REG_TICK_ENABLE:
				{
					packetData = SendUnsignedType(data);
					break;
				}
				
				case Register::REG_TICK_CURRENT:
				{
					packetData = SendUshortFloatType(data, FACTOR_CURRENT);
					break;
				}
				
				case Register::REG_TICK_ANGLE_CW:
				{
					packetData = SendUshortFloatType(data, FACTOR_ANGLE);
					break;
				}
				
				case Register::REG_TICK_DURATION_MIN:
				case Register::REG_TICK_DURATION_MAX:
				{
					packetData = SendUshortFloatType(data, FACTOR_DURATION);
					break;
				}

				case Register::REG_TICK_VELOCITY_FACTOR:
				{
					packetData = SendUshortFloatType(data, FACTOR_VELOCITY_FACTOR);
					break;
				}

				case Register::REG_TICK_INDEX_WINDOW:
				{
					packetData = SendUshortFloatType(data, FACTOR_WINDOW);
					break;
				}

				case Register::REG_TICK_STICKINESS_PREVENTION_FACTOR:
				{
					packetData = SendUshortFloatType(data, FACTOR_STICKINESS);
					break;
				}

				case Register::REG_TICK_ANGLE_CCW:
				{
					packetData = SendUshortFloatType(data, FACTOR_ANGLE);
					break;
				}

				case Register::REG_TICK_ACTIVE_DIRECTION:
				{
					packetData = SendUnsignedType(data);
					break;
				}

				case Register::REG_TICK_FREEWHEELING_VELOCITY_THRESHOLD:
				{
					packetData = SendUshortFloatType(data, FACTOR_VELOCITY_THRESHOLD);
					break;
				}

				case Register::REG_TICK_FREEWHEELING_EXTENSION_TIME:
				{
					packetData = SendUshortFloatType(data, FACTOR_DURATION);
					break;
				}

				case Register::REG_TICK_WINDOW:
				{
					packetData = SendUshortFloatType(data, FACTOR_WINDOW);
					break;
				}

				case Register::REG_TICK_START_ANGLE:
				case Register::REG_TICK_STOP_ANGLE:
				{
					packetData = SendUshortFloatType(data, FACTOR_ANGLE);
					break;
				}

				// Haptic Function Barrier
				case Register::REG_BARRIER_ENABLE:
				{
					packetData = SendUnsignedType(data);
					break;
				}

				case Register::REG_BARRIER_CURRENT:
				{
					packetData = SendUshortFloatType(data, FACTOR_CURRENT);
					break;
				}

				case Register::REG_BARRIER_START_ANGLE:
				case Register::REG_BARRIER_STOP_ANGLE:
				{
					packetData = SendUshortFloatType(data, FACTOR_ANGLE);
					break;
				}

				case Register::REG_BARRIER_POLARITY:
				case Register::REG_BARRIER_LIMIT_ANGLE:
				{
					packetData = SendUnsignedType(data);
					break;
				}

				// Haptic Function Current
				case Register::REG_CURRENT_ENABLE:
				{
					packetData = SendUnsignedType(data);
					break;
				}

				case Register::REG_CURRENT_START_ANGLE:
				case Register::REG_CURRENT_STOP_ANGLE:
				{
					packetData = SendUshortFloatType(data, FACTOR_ANGLE);
					break;
				}

				case Register::REG_CURRENT_START_CURRENT_CW:
				case Register::REG_CURRENT_STOP_CURRENT_CW:
				case Register::REG_CURRENT_START_CURRENT_CCW:
				case Register::REG_CURRENT_STOP_CURRENT_CCW:
				{
					packetData = SendUshortFloatType(data, FACTOR_CURRENT);
					break;
				}

				case Register::REG_CURRENT_FREEWHEELING_VELOCITY_THRESHOLD:
				{
					packetData = SendUshortFloatType(data, FACTOR_VELOCITY_THRESHOLD);
					break;
				}

				case Register::REG_CURRENT_FREEWHEELING_EXTENSION_TIME:
				{
					packetData = SendUshortFloatType(data, FACTOR_DURATION);
					break;
				}

				case Register::REG_CURRENT_ACTIVE_DIRECTION:
				{
					packetData = SendUnsignedType(data);
					break;
				}

				case Register::REG_CURRENT_STREAMING_ENABLE:
				{
					packetData = SendBytesType(data);
					break;
				}

				case Register::REG_CURRENT_STREAMING_CURRENT:
				{
					packetData = SendUshortFloatType(data, FACTOR_CURRENT);
					break;
				}

				// Haptic Function Torque
				case Register::REG_TORQUE_ENABLE:
				{
					packetData = SendUnsignedType(data);
					break;
				}

				case Register::REG_TORQUE_START_ANGLE:
				case Register::REG_TORQUE_STOP_ANGLE:
				{
					packetData = SendUshortFloatType(data, FACTOR_ANGLE);
					break;
				}

				case Register::REG_TORQUE_START_FACTOR_CW:
				case Register::REG_TORQUE_STOP_FACTOR_CW:
				case Register::REG_TORQUE_START_FACTOR_CCW:
				case Register::REG_TORQUE_STOP_FACTOR_CCW:
				{
					packetData = SendUshortFloatType(data, FACTOR_TORQUE);
					break;
				}

				case Register::REG_TORQUE_FREEWHEELING_VELOCITY_THRESHOLD:
				{
					packetData = SendUshortFloatType(data, FACTOR_VELOCITY_THRESHOLD);
					break;
				}

				case Register::REG_TORQUE_FREEWHEELING_EXTENSION_TIME:
				{
					packetData = SendUshortFloatType(data, FACTOR_DURATION);
					break;
				}

				case Register::REG_TORQUE_ACTIVE_DIRECTION:
				{
					packetData = SendUnsignedType(data);
					break;
				}

				// Haptic Function Lock
				case Register::REG_LOCK_ENABLE:
				case Register::REG_LOCK_DIRECTION:
				{
					packetData = SendUnsignedType(data);
					break;
				}

				case Register::REG_LOCK_CURRENT:
				{
					packetData = SendUshortFloatType(data, FACTOR_CURRENT);
					break;
				}

				// Haptic Function Single Tick
				case Register::REG_SINGLE_TICK_ENABLE:
				case Register::REG_SINGLE_TICK_ACTIVE_DIRECTION:
				case Register::REG_SINGLE_TICK_MODE:
				{
					packetData = SendUnsignedType(data);
					break;
				}

				case Register::REG_SINGLE_TICK_ANGLE:
				{
					packetData = SendUshortFloatType(data, FACTOR_ANGLE);
					break;
				}
				
				case Register::REG_SINGLE_TICK_CURRENT:
				{
					packetData = SendUshortFloatType(data, FACTOR_CURRENT);
					break;
				}

				case Register::REG_SINGLE_TICK_DURATION_MIN:
				case Register::REG_SINGLE_TICK_DURATION_MAX:
				{
					packetData = SendUshortFloatType(data, FACTOR_DURATION);
					break;
				}

				case Register::REG_SINGLE_TICK_WINDOW:
				{
					packetData = SendUshortFloatType(data, FACTOR_WINDOW);
					break;
				}

				case Register::REG_SINGLE_TICK_VELOCITY_FACTOR:
				{
					packetData = SendUshortFloatType(data, FACTOR_VELOCITY_FACTOR);
					break;
				}

				// Haptic Function Haptics Generator
				case Register::REG_HAPTICS_GENERATOR_ENABLE:
				{
					packetData = SendUnsignedType(data);
					break;
				}

				// Custom Mode
				case Register::REG_CUSTOM_MODE_ENABLE:
				case Register::REG_CUSTOM_MODE:
				{
					packetData = SendUnsignedType(data);
					break;
				}

				case Register::REG_CUSTOM_PARAMETER1:
				{
					packetData = SendUshortFloatType(data, customParamFactor[0]);
					break;
				}

				case Register::REG_CUSTOM_PARAMETER2:
				{
					packetData = SendUshortFloatType(data, customParamFactor[1]);
					break;
				}

				case Register::REG_CUSTOM_PARAMETER3:
				{
					packetData = SendUshortFloatType(data, customParamFactor[2]);
					break;
				}

				case Register::REG_CUSTOM_PARAMETER4:
				{
					packetData = SendUshortFloatType(data, customParamFactor[3]);
					break;
				}

				case Register::REG_CUSTOM_PARAMETER5:
				{
					packetData = SendUshortFloatType(data, customParamFactor[4]);
					break;
				}

				case Register::REG_CUSTOM_PARAMETER6:
				{
					packetData = SendUshortFloatType(data, customParamFactor[5]);
					break;
				}

				case Register::REG_CUSTOM_PARAMETER7:
				{
					packetData = SendUshortFloatType(data, customParamFactor[6]);
					break;
				}

				case Register::REG_CUSTOM_PARAMETER1_PROPERTIES:
				case Register::REG_CUSTOM_PARAMETER2_PROPERTIES:
				case Register::REG_CUSTOM_PARAMETER3_PROPERTIES:
				case Register::REG_CUSTOM_PARAMETER4_PROPERTIES:
				case Register::REG_CUSTOM_PARAMETER5_PROPERTIES:
				case Register::REG_CUSTOM_PARAMETER6_PROPERTIES:
				case Register::REG_CUSTOM_PARAMETER7_PROPERTIES:
				{
					packetData = SendUnsignedType(data);
					break;
				}

				default:
				{
					return;
				}
			}

			SerialWrite(reg, *(packetData + 1), *packetData);
			
			delete packetData;
			packetData = nullptr;
		}
	}
}
